# -*- coding: utf-8 -*-
"""Test IRDB compliance."""

import os
import tempfile

import pytest
import scopesim
import yaml
from scopesim import rc
from scopesim.effects.data_container import DataContainer

from micado.hardware.DetectorArray import DetectorArray
from micado.main.Hardware import Hardware
from micado.main.MObject import MObject


# TODO: Refactor this test, currently a bit messy.
# TODO: Generalize to other hardware once implemented.


def verify_hardware(hardware, yam):
    """Verify the attributes of hardware.

    The should be the same as the kwargs/properties in yaml,
    including default values.
    """
    assert issubclass(hardware, Hardware)

    # The attributes and defaults in MicadoWISE
    mkwargs = {
        a: getattr(hardware, a).default
        for a in hardware.get_persistent_properties()
        if not a.startswith("dummy")
        and a not in ["object_id", "filename", "globalname"]
        and (
            # Annoying tuple check because prop_type can be (int, float)
            isinstance(getattr(hardware, a).prop_type, tuple)
            or not issubclass(getattr(hardware, a).prop_type, MObject)
        )
    }

    # HardwareThings have properties and effects have kwargs.
    # They should not have both but can have neither.
    assert not ("kwargs" in yam and "properties" in yam)
    key_kw = "properties" if "properties" in yam else "kwargs"

    # The attributes and defaults in the IRDB
    if set(yam.get(key_kw, {}).keys()) == {"filename"}:
        dc = DataContainer(filename=yam[key_kw]["filename"])
        yam2 = {
            # Any value in the table is allowed as default.
            k: set(v.data)
            for k, v in dc.table.columns.items()
        }
    elif key_kw in yam:
        yam2 = {
            k: {v}
            for k, v in yam[key_kw].items()
            # Starting with an exclamation mark are defined elsewhere.
            if not (isinstance(v, str) and v.startswith("!"))
        }
    else:
        # No attributes in IRDB. Just use the ones in MicadoWISE
        # TODO: do something smarter?
        yam2 = {k: {v} for k, v in mkwargs.items()}

    attrs_extra = mkwargs.keys() - yam2.keys()
    attrs_missing = yam2.keys() - mkwargs.keys()
    # TODO: fix these x_len and y_len
    attrs_extra_expected = {"name"}
    attrs_missing_expected = {"x_len", "y_len"}

    assert attrs_extra - attrs_extra_expected == set()
    assert attrs_missing - attrs_missing_expected == set()

    for k in mkwargs:
        if k not in ["id", "name"]:
            assert mkwargs[k] in yam2[k]


@pytest.mark.xfail(reason="TODO: Fix IRDB compliance")
def test_irdbcompliance():
    """Test IRDB compliance."""
    with tempfile.TemporaryDirectory() as tempdir:
        tdn = tempdir

        rc.__config__["!SIM.file.local_packages_path"] = os.path.abspath(tdn)
        rc.__search_path__.insert(0, os.path.abspath(tdn + "/MICADO"))
        scopesim.server.database.download_package(
            "instruments/MICADO.zip", save_dir=tdn
        )
        # TODO: get from default.yaml
        ydet = yaml.safe_load(open(tempdir + "/MICADO/MICADO_H4RG.yaml"))
        print(ydet)
        assert ydet["object"] == "detector"
        # TODO: generalize for other hardware
        hardware = DetectorArray

        attrs_effects = [
            a
            for a in hardware.get_persistent_properties()
            # Annoying tuple check because prop_type can be (int, float)
            if not isinstance(getattr(hardware, a).prop_type, tuple)
            and issubclass(getattr(hardware, a).prop_type, Hardware)
        ]
        eff_names = [e["name"] for e in ydet["effects"]]
        attrs_extra = set(attrs_effects) - set(eff_names)
        attrs_missing = set(eff_names) - set(attrs_effects)
        attrs_missing_expected = {
            "SimpleCADO_1024_detector_window",
            "detector_linearity",
            "exposure_action",
            "qe_curve",
            "readout_noise",
        }
        assert attrs_extra == set()
        assert attrs_missing == attrs_missing_expected
        verify_hardware(DetectorArray, ydet)
        for attr_e in attrs_effects:
            effect = getattr(hardware, attr_e).prop_type
            yam = [e for e in ydet["effects"] if e["name"] == attr_e][0]
            verify_hardware(effect, yam)

        # TODO: Assert properties
        print(ydet["properties"])
